/* imsub.c */
/* subroutine to treat image data */
/* 93/01/20 I.Yamamura */
/* 93/06/12 renamed as imsub.c */
/* 94/04/21 some modifications on Arith() */
/* 94/07/02 some modifications on Arith() */
/* 03/01/04 malloc.h -> stdlib.h */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <memory.h>

#include "macro.h"
#include "funcs.h"

#include "fits.h"

#ifndef DBL_MAX
#  define DBL_MAX 1.7976931348623157e308 /* same as HP */
#endif

/* internal subroutine definition */
long *CutImage();

/* ======================================================================== */
void AddPlane(img, f, p1, p2, img2, f2, av)
/* add p1-th to p2-th plane of img and return to img2 */
   long *img;
   fits_header *f;
   int p1, p2;
   long **img2;
   fits_header *f2;
   int av;  /* if (av == 1) then average intensity is returned. */
{
   int ic, i;
   double *dimg;
   long *limg;
   double np;
   int size;
   double dmax, dmin;

   if ((p1 < 0) || (p2 >= f->naxis3)) {
      puts("Plane number is out of range.");
      return;
   }

   size = f->naxis1 * f->naxis2;
   dimg = (double *)calloc(size, sizeof(double));
   for (ic = p1; ic <= p2; ic++) {
      limg = img + ic * size;
      for (i = 0; i < size; i++)
         dimg[i] += limg[i] * f->bscale + f->bzero;
   }
   if (av == 1) np = p2 - p1 + 1; else np = 1.0;
   dmax = dmin = dimg[0] / np;
   for (i = 0; i < size; i++) {
      dimg[i] /= np;
      if (dimg[i] > dmax) dmax = dimg[i];
      if (dimg[i] < dmin) dmin = dimg[i];
   }

   *f2 = *f;
   f2->naxis3 = 1;
   f2->datamax = dmax;
   f2->datamin = dmin;
   f2->bscale = (dmax - dmin) / pow(2.0, 32.0);
   f2->bzero = (dmax + dmin) / 2.0;
   f2->crval3 = ((p1 + p2) / 2.0 + 1 - f2->crpix3) * f2->cdelt3 + f2->crval3;
   f2->crpix3 = 1;
   f2->size = size;

   (*img2) = (long *)calloc(size, sizeof(long));
   for (i = 0; i < size; i++)
      (*img2)[i] = (dimg[i] - f2->bzero) / f2->bscale;

   free(dimg);
} /* AddPlane() */

/* ======================================================================== */
void PickPlane(img, f, p1, p2, imgo, fo)
/* pick out from p1-th to p2-th plane of img */
   long *img;
   fits_header *f;
   int p1, p2;
   long **imgo;
   fits_header *fo;
{
   int size;

   if ((p1 < 0) || (p2 >= f->naxis3)) {
      puts("PickPlane : Plane number is out of range.");
      return;
   }

   size = f->naxis1 * f->naxis2;
   *imgo = img + p1 * size;

   *fo = *f;
   fo->naxis3 = (p2 - p1 + 1);
   fo->crval3 = (p1 + 1 - f->crpix3) * f->cdelt3 + f->crval3;
   fo->crpix3 = 1;
   fo->size = f->naxis1 * f->naxis2 * (p2 - p1 + 1);

} /* PickPlane() */

/* ======================================================================== */
long *CutImage(img, f, x1, y1, x2, y2)
/* cut image data from (x1, y1) to (x2, y2) */
/* the origin of the coordinate is (0,0) */
/* have some bug ??? 94/09/08 */
   long *img;       /* original image */
   fits_header *f;  /* original fits header */
   int x1, y1, x2, y2;
{
   int ic, ix, iy;
   int nx = x2 - x1 + 1;
   int ny = y2 - y1 + 1;
   long *cimg;
   unsigned size;

   if ((nx < 0) || (ny < 0)) {
      puts("CutImage(): error! number of column or raw is negative.");
      return NULL;
   }
   size = f->naxis3 * nx * ny;
   cimg = (long *)calloc(size, sizeof(long));

   for (ic = 0; ic < f->naxis3; ic++) {
      for (iy = 0; iy < ny; iy++)
         memcpy(cimg+(ic*ny+iy)*nx,
            img+(ic*f->naxis2+(y1+iy))*f->naxis1+x1, nx*sizeof(long));
   }

   return cimg;
} /* CutImage() */

void Triming(img1, f1, img2, f2, lx, ly, cx, cy)
/* cut image centerd at (cx, cy) and size of (2*lx+1)*(2*ly+1) pixcels. */
/* size definition changed as lx * lx 94/06/18 */
/* center definition changed again. 94/07/07 */
/* the origin of the coordinate is (0,0) */
/* don't call CutImage() subroutine (its included) 94/09/08 */
/* cx, cy < 0 => peak 94/10/05 */
/* do not change output image size even when the region is out of the */
/* original image. 96/02/01 */
   long *img1;       /* original image */
   long **img2;      /* trimed image */
   fits_header *f1;  /* original fits header */
   fits_header *f2;  /* result fits header */
   int lx, ly;       /* size of output image */
   int cx, cy;       /* center of triming */
{
   int x1, y1; /* left bottum corner */
   int x2, y2; /* right top corner */
   int i1, i2, i3;      /* counter */
   unsigned int of2, of1;  /* offset */
   long lmax, lmin;  /* maximum and minimum data */
   double d, dmax;

   cx--; cy--;

   if (cx < 0 || cy < 0) {
      dmax = -99999.0;
      for (i3 = 0; i3 < f1->naxis3; i3++) {
         for (i2 = 0; i2 < f1->naxis2; i2++) {
            for (i1 = 0; i1 < f1->naxis1; i1++) {
               d = img1[(i3 * f1->naxis2 + i2) * f1->naxis1 + i1]
                 * f1->bscale + f1->bzero;
               if (dmax == -99999.0  || d > dmax) {
                  dmax = d; cx = i1; cy = i2;
               }
            }
         }
      }
      fprintf(stderr, "peak is at (%d, %d)\n", cx+1, cy+1);
   }   

   x1 = cx - (lx - 1) / 2;    x2 = x1 + lx - 1;
/*   if (x1 < 0) x1 = 0;  if (x2 > f1->naxis1 - 1) x2 = f1->naxis1 - 1; 96/02/01 */
   y1 = cy - (ly - 1) / 2;    y2 = y1 + ly - 1;
/*   if (y1 < 0) y1 = 0;  if (y2 > f1->naxis2 - 1) y2 = f1->naxis2 - 1; 96/02/01 */

   *f2 = *f1;
   f2->naxis1 = x2 - x1 + 1;
   f2->naxis2 = y2 - y1 + 1;
   f2->crpix1 = cx - x1 + 1;
   f2->crpix2 = cy - y1 + 1;
   f2->crval1 = (cx + 1 - f1->crpix1) * f1->cdelt1 + f1->crval1;
   f2->crval2 = (cy + 1 - f1->crpix2) * f1->cdelt2 + f1->crval2;
   f2->size = f2->naxis1 * f2->naxis2 * f2->naxis3;

   if ((f2->naxis1 < 0) || (f2->naxis2 < 0)) {
      fprintf(stderr, "Triming: error! number of column or raw is negative.");
      exit(0);
   }
   (*img2) = (long *)calloc(f2->size, sizeof(long));

   lmax = lmin = img1[y1 * f1->naxis1 + x1];
   for (i3 = 0; i3 < f2->naxis3; i3++) {
      for (i2 = 0; i2 < f2->naxis2; i2++) {
         for (i1 = 0; i1 < f2->naxis1; i1++) {
            of2 = (i3 * f2->naxis2 + i2) * f2->naxis1 + i1;
            of1 = (i3 * f1->naxis2 + y1 + i2) * f1->naxis1 + x1 + i1;
            if (i2 + y1 >= 0 && i2 + y1 < f1->naxis2 &&
                i1 + x1 >= 0 && i1 + x1 < f1->naxis1) {
               (*img2)[of2] = img1[of1];
               if ((*img2)[of2] > lmax) lmax = (*img2)[of2];
               if ((*img2)[of2] < lmin) lmin = (*img2)[of2];
            } else {
               (*img2)[of2] = f2->blank;
            }
         }
      }
   }

   f2->datamax = lmax * f2->bscale + f2->bzero;
   f2->datamin = lmin * f2->bscale + f2->bzero;

} /* Triming */

/* ======================================================================== */
void Axchange(img, f, axc, img2, f2)
/* exchange two axis 92/01/30 */
   long *img;
   fits_header *f;
   int axc;          /* axis to be conserved. if 0, only copied. */
   long **img2;
   fits_header *f2;
{
   int i, i1, i2, i3;
   int j, j1, j2, j3;

   if ((axc < 0) || (axc > f->naxis)) {
      fprintf(stderr, "Axchange : Invalid axis (axc = %d)\n", axc);
      exit(-1);
   }

   if (axc == 0) {
      *f2 = *f;
      *img2 = img;
      return;
   }
   *f2 = *f;
   if (axc == 3) {
      f2->naxis1 = f->naxis2;          f2->naxis2 = f->naxis1;
      f2->crval1 = f->crval2;          f2->crval2 = f->crval1;
      f2->cdelt1 = f->cdelt2;          f2->cdelt2 = f->cdelt1;
      f2->crpix1 = f->crpix2;          f2->crpix2 = f->crpix1;
      strcpy(f2->ctype1, f->ctype2);   strcpy(f2->ctype2, f->ctype1);
   } else if (axc == 1) {
      f2->naxis2 = f->naxis3;          f2->naxis3 = f->naxis2;
      f2->crval2 = f->crval3;          f2->crval3 = f->crval2;
      f2->cdelt2 = f->cdelt3;          f2->cdelt3 = f->cdelt2;
      f2->crpix2 = f->crpix3;          f2->crpix3 = f->crpix2;
      strcpy(f2->ctype2, f->ctype3);   strcpy(f2->ctype3, f->ctype2);
   } else if (axc == 2) {
      f2->naxis3 = f->naxis1;          f2->naxis1 = f->naxis3;
      f2->crval3 = f->crval1;          f2->crval1 = f->crval3;
      f2->cdelt3 = f->cdelt1;          f2->cdelt1 = f->cdelt3;
      f2->crpix3 = f->crpix1;          f2->crpix1 = f->crpix3;
      strcpy(f2->ctype3, f->ctype1);   strcpy(f2->ctype1, f->ctype3);
   }

   *img2 = (long *)calloc(f->size, sizeof(long));
   for (i = 0; i < f->size; i++) {
      i1 = i % f->naxis1;
      i2 = ((i - i1) / f->naxis1) % f->naxis2;
      i3 = ((i - i1) / f->naxis1) / f->naxis2;
      if (axc == 3) { j1 = i2; j2 = i1; j3 = i3;}
      else if (axc == 1) { j2 = i3; j3 = i2; j1 = i1;}
      else if (axc == 2) { j3 = i1; j1 = i3; j2 = i2;}
      j = j1+f2->naxis1*(j2 + j3*f2->naxis2);
/*printf("%8d%3d%3d%3d%3d%3d%3d%8d\n", i, i1, i2, i3, j1, j2, j3, j);/**/
      (*img2)[j] = img[i];
   }

} /* Axchange() */

/* ======================================================================== */
void FluxRef(f, ref)
/* Flux reference 92/02/26 */
   fits_header *f;
   double ref;
{
   f->bscale *= ref;
   f->bzero  *= ref;
} /* FxRef() */

/* ======================================================================== */
void Arith(img1, f1, img2, f2, imgo, fo, pr, rms1, rms2, sb)
/* Arithmatic calculation of img1 and img2. 93/03/04 */
/* if nuber of plane in img1 is larger than that of img2, */
/* then the last plane of img2 are operated to rest of img1 94/07/02 */
   long *img1, *img2;
   fits_header *f1 ,*f2;
   long **imgo;
   fits_header *fo;
   char *pr;      /* one of [ADD, SUB, MUL, DIV] */
   double rms1, rms2;   /* only good quality data are treated */
   double sb;           /* substitute value for rejected pixcel */
{
/* operation mode list */
#  define ADD 1
#  define SUB 2
#  define MUL 3
#  define DIV 4

   int i, k1, k2;
   int mode;
   int size;
   double *dimg;
   double d1, d2;
   double dmax, dmin;
   double blank;

   if (f1->naxis1 != f2->naxis1 || f1->naxis2 != f2->naxis2) {
      fprintf(stderr, "Arith : Two images are inconpatible size.\n");
      exit(-1);
   }
        if (strcmp(pr, "ADD") == 0) mode = ADD;
   else if (strcmp(pr, "SUB") == 0) mode = SUB;
   else if (strcmp(pr, "MUL") == 0) mode = MUL;
   else if (strcmp(pr, "DIV") == 0) mode = DIV;
   else {
      fprintf(stderr, "Arith : unknown operation mode %s.\n", pr);
      exit(-1);
   }

   size = f1->naxis1 * f1->naxis2;
   blank = f1->blank * f1->bscale + f1->bzero;
   rms1 = fabs(rms1); rms2 = fabs(rms2);
   dimg = (double *)calloc(f1->size, sizeof(double));

   for (k1 = 0; k1 < f1->naxis3; k1++) {
      if (k1 >= f2->naxis3) k2 = f2->naxis3 - 1; else k2 = k1;
      for (i = 0; i < size; i++) {
         d1 = img1[k1 * size + i] * f1->bscale + f1->bzero;
         d2 = img2[k2 * size + i] * f2->bscale + f2->bzero;
         if (fabs(d1) > rms1 && fabs(d2) > rms2) {
                 if (mode == ADD) dimg[k1 * size + i] = d1 + d2;
            else if (mode == SUB) dimg[k1 * size + i] = d1 - d2;
            else if (mode == MUL) dimg[k1 * size + i] = d1 * d2;
            else if (mode == DIV)
               if (d2 == 0)   dimg[k1 * size + i] = DBL_MAX;
               else           dimg[k1 * size + i] = d1 / d2;
         } else {
            dimg[k1 * size + i] = sb;
         }
      }
   }

   dmax = dmin = dimg[0];
   for (i = 0; i < f1->size; i++) {
      if (dimg[i] > dmax) dmax = dimg[i];
      if (dimg[i] < dmin) dmin = dimg[i];
   }

   *fo = *f1;
   fo->datamax = dmax;
   fo->datamin = dmin;
   fo->bscale = (dmax - dmin) / pow(2.0, 32.0);
   fo->bzero = (dmax + dmin) / 2.0;

   (*imgo) = (long *)calloc(f1->size, sizeof(long));
   for (i = 0; i < f1->size; i++)
      (*imgo)[i] = (dimg[i] - fo->bzero) / fo->bscale;

   free(dimg);
} /* Arith() */

/* ======================================================================== */
void Multi(img, f, v1, v2, np, imgo, fo)
/* interpolate planes and reconstruct np planes between v1 and v2 */
/* 93/03/04 I.Yamamura */
   long *img;
   fits_header *f;
   int np;
   double v1, v2;
   long **imgo;
   fits_header *fo;
{
#define IDV ((((iv)*iymax)+iy)*ixmax + ix)
#define IDP ((((ip)*iymax)+iy)*ixmax + ix)

   int i, ix, iy, iv, ip;
   int ixmax = f->naxis1;
   int iymax = f->naxis2;
   int ivmax = f->naxis3;
   int psize = np * f->naxis2 * f->naxis1;

   double *dimg;
   double dmax, dmin;
   double p, fp;
   double *vv, *fx, *ww;

   double prval = v1;
   double pdelt = (v2 - v1) / (double)(np-1);
   double prpix = 1;

   if (f->naxis < 3) {
      fprintf(stderr, "Multi : more than two planes are needed.\n");
      exit(-1);
   }

   dimg = (double *)calloc(psize, sizeof(double));

   vv = (double *)calloc(ivmax+2, sizeof(double));
   fx = (double *)calloc(ivmax+2, sizeof(double));
/*   ww = (double *)calloc(ivmax+2, sizeof(double));/**/

   for (ix = 0; ix < ixmax; ix++) {
      for (iy = 0; iy < iymax; iy++) {
         vv[0] = (-1 + 1 - f->crpix3) * f->cdelt3 + f->crval3;
         fx[0] = 0.0;
         for (iv = 0; iv < ivmax; iv++) {
            vv[iv+1] = (iv + 1 - f->crpix3) * f->cdelt3 + f->crval3;
            fx[iv+1] = img[IDV] * f->bscale + f->bzero;
         }
         vv[ivmax+1] = (ivmax+1 + 1 - f->crpix3) * f->cdelt3 + f->crval3;
         fx[ivmax+1] = 0.0;

/*        make_sp_table(ivmax+2, vv, fx, ww);/**/
         for (ip = 0; ip < np; ip++) {
            p = (ip + 1 - prpix) * pdelt + prval;
/*            fp = spline(p, ivmax+2, vv, fx, ww);/**/
            fp = LinerInterpolate(p, ivmax+2, vv, fx);/**/
            if (isnan(fp)) {
               fprintf(stderr, "Multi : %e at %d %d %d\n", fp, ix, iy, ip);
               fp = 0.0;
            }
            dimg[IDP] = fp;
         }
      }
   }

   free(vv);
   free(fx);
/*   free(ww);/**/
            
   dmax = dmin = dimg[0] / np;
   for (i = 0; i < psize; i++) {
      if (dimg[i] > dmax) dmax = dimg[i];
      if (dimg[i] < dmin) dmin = dimg[i];
   }

   *fo = *f;
   fo->naxis3 = np;
   fo->datamax = dmax;
   fo->datamin = dmin;
   fo->bscale = (dmax - dmin) / pow(2.0, 32.0);
   fo->bzero = (dmax + dmin) / 2.0;
   fo->crval3 = prval;
   fo->cdelt3 = pdelt;
   fo->crpix3 = prpix;
   fo->size = psize;

   (*imgo) = (long *)calloc(psize, sizeof(long));
   for (i = 0; i < psize; i++)
      (*imgo)[i] = (dimg[i] - fo->bzero) / fo->bscale;

   free(dimg);
} /* Multi() */

/* ======================================================================== */
void RMS(img, f, p1, p2, rms)
/* calculate rms from p1-th to p2-th plane */
   long *img;
   fits_header *f;
   int p1, p2;
   double *rms;      /* returned value */
{
   int ip, i;
   int size = f->naxis1 * f->naxis2;
   double sum = 0;

   if ((p1 < 0) || (p2 >= f->naxis3)) {
      fprintf(stderr, "RMS : Plane number is out of range.\n");
      return;
   }
   if (p2 < p1) {
      fprintf(stderr, "RMS : p1 < p2.\n");
      return;
   }

   for (ip = p1; ip <= p2; ip++) {
      for (i = 0; i < size; i++)
         sum += sqr(img[ip*size + i] * f->bscale + f->bzero);
   }
   *rms = sqrt(sum / (double)((p2 - p1 + 1)*size));
} /* RMS() */

/* ======================================================================== */
void Connect(img1, f1, img2, f2, imgo, fo)
/* Connect two image files. img2 follows img1. */
   long *img1, *img2;
   fits_header *f1 ,*f2;
   long **imgo;
   fits_header *fo;
{
   int i, j;
   int size;
   double *dimg;
   double dmax, dmin;

   if (f1->naxis1 != f2->naxis1 || f1->naxis2 != f2->naxis2) {
      fprintf(stderr, "Connect : Two images are inconpatible size.\n");
      exit(-1);
   }

   size = f1->naxis1 * f1->naxis2;
   dimg = (double *)calloc(size*(f1->naxis3 + f2->naxis3), sizeof(double));
   j = 0;
   for (i = 0; i < size*f1->naxis3; i++)
      dimg[j++] = img1[i] * f1->bscale + f1->bzero;
   for (i = 0; i < size*f2->naxis3; i++)
      dimg[j++] = img2[i] * f2->bscale + f2->bzero;

   size = size * (f1->naxis3 + f2->naxis3);
   dmax = dmin = dimg[0];
   for (i = 0; i < size; i++) {
      if (dimg[i] > dmax) dmax = dimg[i];
      if (dimg[i] < dmin) dmin = dimg[i];
   }

   *fo = *f1;
   fo->naxis3 = f1->naxis3 + f2->naxis3;
   fo->datamax = dmax;
   fo->datamin = dmin;
   fo->bscale = (dmax - dmin) / pow(2.0, 32.0);
   fo->bzero = (dmax + dmin) / 2.0;
   fo->size = size;

   (*imgo) = (long *)calloc(size, sizeof(long));
   for (i = 0; i < size; i++)
      (*imgo)[i] = (dimg[i] - fo->bzero) / fo->bscale;

   free(dimg);
} /* Connect() */

/* end of fitsub.c */
