/* rwfits.c */
/* read FITS format file (written by AIPS) */
/* 93/01/13 I.Yamamura */
/* writing in FITS format routines are added */
/* 93/01/19 I.Yamamura */
/* 93/03/21 WriteImage: force to write mode added. */
/* 93/04/14 add "size" in fits_header and can treat one-plaine image. */
/* 93/06/12 renamed as imrw.c */
/*          extended fits */
/* 94/04/14 add "blank" */
/* 94/08/27 incorrect input header (no datamin and datamax) can be treated */
/* 94/08/27 can read 8-byte double data (temporary) */
/* 95/03/10 extended to treat UV data */

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <sys/fcntl.h>
#include <errno.h>
#include <sys/errno.h>

#include "macro.h"

#include "fits.h"

/* ======================================================================== */
#define KeyWord(s, kw) strncmp(kw, s, 8)
#define PKeyWord(s, kw) strncmp(kw, s, 5)
#define ReadBln(s, c) c = s[29]
#define ReadDbl(s, d) sscanf(&s[10], "%lf", &(d))
#define ReadInt(s, i) sscanf(&s[10], "%d", &(i))
void    ReadStr(s, ss)
   char *s, *ss;
{
   int i = 9;
   int j = 0;

   while (s[i++] != '\'');
   while ((s[i] != '\'') && (j < 8)) ss[j++] = s[i++];
   ss[j] = '\0';
} /* ReadStr() */

void InitFitsHeader(f)
/* initialize fits header */
/* 93/06/12 I.Yamamura */
   fits_header *f;
{
   f->simple  = 'F';
   f->bitpix  = 0;
   f->naxis   = 0;
   f->naxis1  = 1;
   f->naxis2  = 1;
   f->naxis3  = 1;
   f->naxis4  = 1;
   f->naxis5  = 0;
   f->naxis6  = 0;
   f->extend  = 'F';
   f->blocked = 'F';
   strcpy(f->object,   "");
   strcpy(f->telescop, "");
   strcpy(f->instrume, "");
   strcpy(f->observer, "");
   strcpy(f->origin,   "");
   strcpy(f->date_obs, "");
   strcpy(f->date_map, "");
   f->bscale  = 1.0;
   f->bzero   = 0.0;
   strcpy(f->bunit, "");
   f->epoch   = 0.0;
   f->blank   = 0;
   f->velref  = 0;
   f->altrval = 0.0;
   f->altrpix = 0.0;
   f->obsra   = -999.0;
   f->obsdec  = -999.0;
   f->restfreq= 0.0;
   f->yshift  = 0.0;
   f->datamax = 0.0;
   f->datamin = 0.0;

   strcpy(f->ctype1, "");
   f->crval1 = 0.0;
   f->cdelt1 = 0.0;
   f->crpix1 = 0.0;
   f->crota1 = 0.0;
   
   strcpy(f->ctype2, "");
   f->crval2 = 0.0;
   f->cdelt2 = 0.0;
   f->crpix2 = 0.0;
   f->crota2 = 0.0;
   
   strcpy(f->ctype3, "");
   f->crval3 = 0.0;
   f->cdelt3 = 0.0;
   f->crpix3 = 0.0;
   f->crota3 = 0.0;
   
   strcpy(f->ctype4, "");
   f->crval4 = 0.0;
   f->cdelt4 = 0.0;
   f->crpix4 = 0.0;
   f->crota4 = 0.0;

   strcpy(f->ctype5, "");
   f->crval5 = 0.0;
   f->cdelt5 = 0.0;
   f->crpix5 = 0.0;
   f->crota5 = 0.0;

   strcpy(f->ctype6, "");
   f->crval6 = 0.0;
   f->cdelt6 = 0.0;
   f->crpix6 = 0.0;
   f->crota6 = 0.0;

/* extended fits */
   f->groups  = 'F';
   f->pcount  = 0;
   f->gcount  = 0;
   f->p = NULL;

/* weight scale for UV data */
   f->wtscal = 1.0;

/* following are inner valiables */
   f->size = 1;

}  /* InitFitsHeader() */

void ReadFitsHeader(infile, f)
/* read FITS header. 93/01/13 I.Yamamura */
/* if P* variables are used, refer the routine for newstar observation. */
   int infile;
   fits_header *f;
{
   char s[81], kw[9];
   char dummy[5];
   int nline = 0;
   
   InitFitsHeader(f);
   s[80] = kw[9] = '\0';
   do {
      nline++;
      read(infile, s, 80); s[80] = '\0';
      if (nline == 1 && KeyWord(s, "SIMPLE  ") != 0) {
         fprintf(stderr, "ReadFitsHeader: This file is not FITS file.\n");
         exit(-1);
      }
      else if (KeyWord(s, "SIMPLE  ") == 0) { ReadBln(s, f->simple); }
      else if (KeyWord(s, "BITPIX  ") == 0) { ReadInt(s, f->bitpix); }
      else if (KeyWord(s, "NAXIS   ") == 0) { ReadInt(s, f->naxis); }
      else if (KeyWord(s, "NAXIS1  ") == 0) { ReadInt(s, f->naxis1); }
      else if (KeyWord(s, "NAXIS2  ") == 0) { ReadInt(s, f->naxis2); }
      else if (KeyWord(s, "NAXIS3  ") == 0) { ReadInt(s, f->naxis3); }
      else if (KeyWord(s, "NAXIS4  ") == 0) { ReadInt(s, f->naxis4); }
      else if (KeyWord(s, "NAXIS5  ") == 0) { ReadInt(s, f->naxis5); }
      else if (KeyWord(s, "NAXIS6  ") == 0) { ReadInt(s, f->naxis6); }
      else if (KeyWord(s, "EXTEND  ") == 0) { ReadBln(s, f->extend); }
      else if (KeyWord(s, "BLOCKED ") == 0) { ReadBln(s, f->blocked); }
      else if (KeyWord(s, "OBJECT  ") == 0) { ReadStr(s, f->object); }
      else if (KeyWord(s, "TELESCOP") == 0) { ReadStr(s, f->telescop); }
      else if (KeyWord(s, "INSTRUME") == 0) { ReadStr(s, f->instrume); }
      else if (KeyWord(s, "OBSERVER") == 0) { ReadStr(s, f->observer); }
      else if (KeyWord(s, "ORIGIN  ") == 0) { ReadStr(s, f->origin); }
      else if (KeyWord(s, "DATE-OBS") == 0) { ReadStr(s, f->date_obs); }
      else if (KeyWord(s, "DATE-MAP") == 0) { ReadStr(s, f->date_map); }
      else if (KeyWord(s, "BSCALE  ") == 0) { ReadDbl(s, f->bscale); }
      else if (KeyWord(s, "BZERO   ") == 0) { ReadDbl(s, f->bzero); }
      else if (KeyWord(s, "BUNIT   ") == 0) { ReadStr(s, f->bunit); }
      else if (KeyWord(s, "EPOCH   ") == 0) { ReadDbl(s, f->epoch); }
      else if (KeyWord(s, "BLANK   ") == 0) { ReadInt(s, f->blank); }
      else if (KeyWord(s, "VELREF  ") == 0) { ReadInt(s, f->velref); }
      else if (KeyWord(s, "ALTRVAL ") == 0) { ReadDbl(s, f->altrval); }
      else if (KeyWord(s, "ALTRPIX ") == 0) { ReadDbl(s, f->altrpix); }
      else if (KeyWord(s, "OBSRA   ") == 0) { ReadDbl(s, f->obsra); }
      else if (KeyWord(s, "OBSDEC  ") == 0) { ReadDbl(s, f->obsdec); }
      else if (KeyWord(s, "RESTFREQ") == 0) { ReadDbl(s, f->restfreq); }
      else if (KeyWord(s, "YSHIFT  ") == 0) { ReadDbl(s, f->yshift); }
      else if (KeyWord(s, "DATAMAX ") == 0) { ReadDbl(s, f->datamax); }
      else if (KeyWord(s, "DATAMIN ") == 0) { ReadDbl(s, f->datamin); }
      else if (KeyWord(s, "CTYPE1  ") == 0) { ReadStr(s, f->ctype1); }
      else if (KeyWord(s, "CRVAL1  ") == 0) { ReadDbl(s, f->crval1); }
      else if (KeyWord(s, "CDELT1  ") == 0) { ReadDbl(s, f->cdelt1); }
      else if (KeyWord(s, "CRPIX1  ") == 0) { ReadDbl(s, f->crpix1); }
      else if (KeyWord(s, "CROTA1  ") == 0) { ReadDbl(s, f->crota1); }
      else if (KeyWord(s, "CTYPE2  ") == 0) { ReadStr(s, f->ctype2); }
      else if (KeyWord(s, "CRVAL2  ") == 0) { ReadDbl(s, f->crval2); }
      else if (KeyWord(s, "CDELT2  ") == 0) { ReadDbl(s, f->cdelt2); }
      else if (KeyWord(s, "CRPIX2  ") == 0) { ReadDbl(s, f->crpix2); }
      else if (KeyWord(s, "CROTA2  ") == 0) { ReadDbl(s, f->crota2); }
      else if (KeyWord(s, "CTYPE3  ") == 0) { ReadStr(s, f->ctype3); }
      else if (KeyWord(s, "CRVAL3  ") == 0) { ReadDbl(s, f->crval3); }
      else if (KeyWord(s, "CDELT3  ") == 0) { ReadDbl(s, f->cdelt3); }
      else if (KeyWord(s, "CRPIX3  ") == 0) { ReadDbl(s, f->crpix3); }
      else if (KeyWord(s, "CROTA3  ") == 0) { ReadDbl(s, f->crota3); }
      else if (KeyWord(s, "CTYPE4  ") == 0) { ReadStr(s, f->ctype4); }
      else if (KeyWord(s, "CRVAL4  ") == 0) { ReadDbl(s, f->crval4); }
      else if (KeyWord(s, "CDELT4  ") == 0) { ReadDbl(s, f->cdelt4); }
      else if (KeyWord(s, "CRPIX4  ") == 0) { ReadDbl(s, f->crpix4); }
      else if (KeyWord(s, "CROTA4  ") == 0) { ReadDbl(s, f->crota4); }
      else if (KeyWord(s, "CTYPE5  ") == 0) { ReadStr(s, f->ctype5); }
      else if (KeyWord(s, "CRVAL5  ") == 0) { ReadDbl(s, f->crval5); }
      else if (KeyWord(s, "CDELT5  ") == 0) { ReadDbl(s, f->cdelt5); }
      else if (KeyWord(s, "CRPIX5  ") == 0) { ReadDbl(s, f->crpix5); }
      else if (KeyWord(s, "CROTA5  ") == 0) { ReadDbl(s, f->crota5); }
      else if (KeyWord(s, "CTYPE6  ") == 0) { ReadStr(s, f->ctype6); }
      else if (KeyWord(s, "CRVAL6  ") == 0) { ReadDbl(s, f->crval6); }
      else if (KeyWord(s, "CDELT6  ") == 0) { ReadDbl(s, f->cdelt6); }
      else if (KeyWord(s, "CRPIX6  ") == 0) { ReadDbl(s, f->crpix6); }
      else if (KeyWord(s, "CROTA6  ") == 0) { ReadDbl(s, f->crota6); }
      else if (KeyWord(s, "GROUPS  ") == 0) { ReadBln(s, f->groups); }
      else if (KeyWord(s, "GCOUNT  ") == 0) { ReadInt(s, f->gcount); }
      else if (KeyWord(s, "PCOUNT  ") == 0) { ReadInt(s, f->pcount);
         if (f->groups == 'T')
            f->p = (p_struct *)calloc(f->pcount, sizeof(p_struct)); }
      else if (PKeyWord(s, "PTYPE") == 0) {
         if (f->groups == 'T')
            ReadStr(s, (f->p[atoi(strncpy(dummy, &(s[5]),3))-1]).type); }
      else if (PKeyWord(s, "PSCAL") == 0) {
         if (f->groups == 'T')
            ReadDbl(s, (f->p[atoi(strncpy(dummy, &(s[5]),3))-1]).scal); }
      else if (PKeyWord(s, "PZERO") == 0) {
         if (f->groups == 'T')
            ReadDbl(s, (f->p[atoi(strncpy(dummy, &(s[5]),3))-1]).zero); }
      else if (strncmp(s, "HISTORY AIPS WTSCAL", 19) == 0) {
         sscanf(&s[21], "%lf", &(f->wtscal));
      }
   } while (KeyWord(s, "END     ") != 0);

   for (; nline % 36 != 0; nline++)
      read(infile, s, 80);

/* check the variables and set inner valiables */
   if (f->naxis > 6) fprintf(stderr, "ReadFitsHeader : NAXIS exceeds 6.\n");

} /* ReadFitsHeader() */

void PrintHeader(f)
/* print header variables for checking */
fits_header *f;
{
   int i;

   printf("SIMPLE  = %c\n", f->simple);
   printf("BITPIX  = %d\n", f->bitpix);
   printf("NAXIS   = %d\n", f->naxis);
   printf("NAXIS1  = %d\n", f->naxis1);
   printf("NAXIS2  = %d\n", f->naxis2);
   printf("NAXIS3  = %d\n", f->naxis3);
   printf("NAXIS4  = %d\n", f->naxis4);
   printf("NAXIS5  = %d\n", f->naxis5);
   printf("NAXIS6  = %d\n", f->naxis6);
   printf("EXTEND  = %c\n", f->extend);
   printf("BLOCKED = %c\n", f->blocked);
   printf("OBJECT  = %s\n", f->object);
   printf("TELESCOP= %s\n", f->telescop);
   printf("INSTRUME= %s\n", f->instrume);
   printf("OBSERVER= %s\n", f->observer);
   printf("ORIGIN  = %s\n", f->origin);
   printf("DATE-OBS= %s\n", f->date_obs);
   printf("DATE-MAP= %s\n", f->date_map);
   printf("BSCALE  = %e\n", f->bscale);
   printf("BZERO   = %e\n", f->bzero);
   printf("BUNIT   = %s\n", f->bunit);
   printf("EPOCH   = %e\n", f->epoch);
   printf("BLANK   = %d\n", f->blank);
   printf("VELREF  = %d\n", f->velref);
   printf("ALTRVAL = %e\n", f->altrval);
   printf("ALTRPIX = %e\n", f->altrpix);
   printf("OBSRA   = %e\n", f->obsra);
   printf("OBSDEC  = %e\n", f->obsdec);
   printf("RESTFREQ= %e\n", f->restfreq);
   printf("YSHIFT  = %e\n", f->yshift);
   printf("DATAMAX = %e\n", f->datamax);
   printf("DATAMIN = %e\n", f->datamin);
   printf("CTYPE1  = %s\n", f->ctype1);
   printf("CRVAL1  = %e\n", f->crval1);
   printf("CDELT1  = %e\n", f->cdelt1);
   printf("CRPIX1  = %e\n", f->crpix1);
   printf("CROTA1  = %e\n", f->crota1);
   printf("CTYPE2  = %s\n", f->ctype2);
   printf("CRVAL2  = %e\n", f->crval2);
   printf("CDELT2  = %e\n", f->cdelt2);
   printf("CRPIX2  = %e\n", f->crpix2);
   printf("CROTA2  = %e\n", f->crota2);
   printf("CTYPE3  = %s\n", f->ctype3);
   printf("CRVAL3  = %e\n", f->crval3);
   printf("CDELT3  = %e\n", f->cdelt3);
   printf("CRPIX3  = %e\n", f->crpix3);
   printf("CROTA3  = %e\n", f->crota3);
   printf("CTYPE4  = %s\n", f->ctype4);
   printf("CRVAL4  = %e\n", f->crval4);
   printf("CDELT4  = %e\n", f->cdelt4);
   printf("CRPIX4  = %e\n", f->crpix4);
   printf("CROTA4  = %e\n", f->crota4);
   printf("GROUPS  = %c\n", f->groups);
   printf("GCOUNT  = %d\n", f->gcount);
   printf("PCOUNT  = %d\n", f->pcount);
   if (f->groups == 'T') {
      for (i = 0; i < f->pcount; i++) {
         printf("PTYPE%-3d  = \"%-8s\", PSCAL%-3d  = %e, PZERO%-3d  = %e\n",
            i+1, (f->p[i]).type, i+1, (f->p[i].scal), (f->p[i].zero));
      }
   }

} /* PrintHeader() */

void ReadImage(fn, img, f)
/* read image data from fits file */
/* double img[NAXIS3*NAXIS2*NAXIS1] is made and returned */
   char *fn;
   long **img;
   fits_header *f;
{
   int infile;
   int i, iv, ira, idec;
   short  *sh;
   double *dimg;
   float  *fimg;
   long size;
   double d;

   infile = open(fn, O_RDONLY, 0);
   if (infile == -1) {
      fprintf(stderr, "ReadImage: file %s does not exist.\n", fn);
      exit(-1);
   }
   ReadFitsHeader(infile, f);
/*   PrintHeader(f);/**/

   if (f->naxis >  4) {
      fprintf(stderr, "ReadImage: NAXIS must be <=4.\n");
      exit(-1);
   }

   if (f->naxis >= 1) size  = f->naxis1;
   if (f->naxis >= 2) size *= f->naxis2;
   if (f->naxis >= 3) size *= f->naxis3;
   if (f->naxis >= 4) size *= f->naxis4;
   (*img) = (long *)calloc((unsigned)size, (unsigned)sizeof(long));

/* if each pixcel is represented by 16bit, force to be 32bit */
   if (f->bitpix == 16) {
      sh = (short *)calloc((unsigned)size, (unsigned)sizeof(short));
      read(infile, sh, size * sizeof(short));
      for (i = 0; i < size; i++)
         (*img)[i] = sh[i];
      f->bitpix = 32;
   } else if (f->bitpix == 32) {  /* normal case, 32bit / pix */
      read(infile, (*img), size * sizeof(long));
/* following codes are temporary 94/08/27 */
   } else if (f->bitpix == -64) { /* 8-byte double / pix */
      dimg = (double *)calloc((unsigned)size, (unsigned)sizeof(double));
      read(infile, dimg, size * sizeof(double));
      if (f->datamax == f->datamin && f->datamax == 0.0) {
         f->datamax = f->datamin = dimg[0];
         for (i = 1; i < size; i++) {
            if (dimg[i] > f->datamax) f->datamax = dimg[i];
            if (dimg[i] < f->datamin) f->datamin = dimg[i];
         }
      }
      f->bscale = 2.0 * (f->datamax - f->datamin) / pow(2.0, 32.0);
      f->bzero = (f->datamax + f->datamin) / 2.0;
      for (i = 0; i < size; i++)
         (*img)[i] = (dimg[i] - f->bzero) / f->bscale;
      f->bitpix = 32;
      free(dimg);
   } else if (f->bitpix == -32) { /* 4-byte double / pix, 94/12/09 */
      fimg = (float *)calloc((unsigned)size, (unsigned)sizeof(float));
      read(infile, fimg, size * sizeof(float));
      if (f->datamax == f->datamin && f->datamax == 0.0) {
         f->datamax = f->datamin = fimg[0];
         for (i = 1; i < size; i++) {
            if (fimg[i] > f->datamax) f->datamax = fimg[i];
            if (fimg[i] < f->datamin) f->datamin = fimg[i];
         }
      }
      f->bscale = 2.0 * (f->datamax - f->datamin) / pow(2.0, 32.0);
      f->bzero = (f->datamax + f->datamin) / 2.0;
      for (i = 0; i < size; i++)
         (*img)[i] = (fimg[i] - f->bzero) / f->bscale;
      f->bitpix = 32;
      free(fimg);
   }
   
/* 94/08/27 */
   if (f->datamax == f->datamin && f->datamax == 0.0) {
      f->datamax = f->datamin = (*img)[0] * f->bscale + f->bzero;
      for (i = 1; i < size; i++) {
         d = (*img)[i] * f->bscale + f->bzero;
         if (d > f->datamax) f->datamax = d;
         if (d < f->datamin) f->datamin = d;
      }
   }

   close(infile);

   f->size = size;

} /* ReadImage() */

/* ----------------------------------------------------------------------- */
void ReadSpectra(fn, spc, f)
/* read spectral data from fits file (newstar spfits file) */
   char *fn;
   long **spc;
   fits_header *f;
{
   int infile;
   int i;
   short *sh;
   long size;

   infile = open(fn, O_RDONLY, 0);
   if (infile == -1) {
      fprintf(stderr, "ReadSpectra: file %s does not exist.\n", fn);
      exit(-1);
   }
   ReadFitsHeader(infile, f);
/*   PrintHeader(f);/**/

   if (f->naxis2 == 0 || f->pcount == 0 ||
         f->gcount == 0 || f->groups != 'T') {
      fprintf(stderr, "ReadSpectra: NAXIS2 == 0 !\n");
      fprintf(stderr, "ReadSpectra: Is (%s) realy newstar spfits?\n", fn);
      exit(-1);
   }
   size = (f->naxis2 + f->pcount) * f->gcount;
   
   (*spc) = (long *)calloc((unsigned)size, (unsigned)sizeof(long));

/* if each pixcel is represented by 16bit, force to be 32bit */
   if (f->bitpix == 16) {
      sh = (short *)calloc((unsigned)size, (unsigned)sizeof(short));
      read(infile, sh, size * sizeof(short));
      for (i = 0; i < size; i++)
         (*spc)[i] = sh[i];
      f->bitpix = 32;
   } else if (f->bitpix == 32)   /* normal case, 32bit / pix */
      read(infile, (*spc), size * sizeof(long));

   close(infile);

   f->size = size;

} /* ReadSpectra() */

/* ----------------------------------------------------------------------- */
void ReadUV(fn, fuv, f)
/* read UV data from fits file (AIPS UV file) */
/* only read header and return the pointer for the data file */
   char *fn;
   int *fuv;        /* pointer for the data */
   fits_header *f;
{
   int infile;

   infile = open(fn, O_RDONLY, 0);
   if (infile == -1) {
      fprintf(stderr, "ReadUV: file %s does not exist.\n", fn);
      exit(-1);
   }
   ReadFitsHeader(infile, f);
/*   PrintHeader(f);/**/

   if (f->bitpix != 32 || f->pcount == 0 ||
       f->gcount == 0 || f->groups != 'T') {
      fprintf(stderr, "ReadUV: Is (%s) realy UV file?\n", fn);
      exit(-1);
   }
   f->size = f->naxis2 * f->naxis3 * f->naxis4 * f->naxis5 * f->naxis6;
   f->size *= f->gcount;
   
   *fuv = infile;

} /* ReadUV() */

/* ======================================================================== */
#define SPC " /                                                "
#define SPC80 "                                                                                " /* empty line */

#define WriteDbl(s, kw, d) { \
   sprintf(s, "%8s= %20.12E%s", kw, d, SPC); \
   write(outfile, s, 80);  \
   nline++; \
}
#define WriteInt(s, kw, i) { \
   sprintf(s, "%8s= %20d%s",    kw, i, SPC); \
   write(outfile, s, 80); \
   nline++; \
}
#define WriteBln(s, kw, b) { \
   sprintf(s, "%8s= %20c%s",    kw, b, SPC); \
   write(outfile, s, 80); \
   nline++; \
}
#define WriteStr(s, kw, ss) { \
   MakeStrHeader(s, kw, ss); \
   write(outfile, s, 80); \
   nline++; \
}
void MakeStrHeader(s, kw, ss)
   char *s, *kw, *ss;
{
   char sf[81];
   int i, j;

   i = j = 0;  sf[i++] = '\'';
   while (ss[j] != '\0') sf[i++] = ss[j++];
   sf[i++] = '\'';
   while(i < 20) sf[i++] = ' ';
   sf[i] = '\0';

   sprintf(s, "%8s= %s%s", kw, sf, SPC);
} /* MakeStrHeader() */

void WriteImHeader(outfile, f)
/* write FITS (image) header by AIPS style. 93/01/13 I.Yamamura */
   int outfile;
   fits_header *f;
{
   char s[81];
   int nline = 0;

   WriteBln(s, "SIMPLE  ", f->simple);
   WriteInt(s, "BITPIX  ", f->bitpix);
   WriteInt(s, "NAXIS   ", f->naxis);
   WriteInt(s, "NAXIS1  ", f->naxis1);
   if (f->naxis >= 2) WriteInt(s, "NAXIS2  ", f->naxis2);
   if (f->naxis >= 3) WriteInt(s, "NAXIS3  ", f->naxis3);
   if (f->naxis >= 4) WriteInt(s, "NAXIS4  ", f->naxis4);
   WriteBln(s, "EXTEND  ", f->extend);
   WriteBln(s, "BLOCKED ", f->blocked);
   WriteStr(s, "OBJECT  ", f->object);
   WriteStr(s, "TELESCOP", f->telescop);
   WriteStr(s, "INSTRUME", f->instrume);
   WriteStr(s, "OBSERVER", f->observer);
   WriteStr(s, "DATE-OBS", f->date_obs);
   WriteStr(s, "DATE-MAP", f->date_map);
   WriteDbl(s, "BSCALE  ", f->bscale);
   WriteDbl(s, "BZERO   ", f->bzero);
   WriteStr(s, "BUNIT   ", f->bunit);
   WriteDbl(s, "EPOCH   ", f->epoch);
   WriteInt(s, "BLANK   ", f->blank);
   WriteInt(s, "VELREF  ", f->velref);
   WriteDbl(s, "ALTRVAL ", f->altrval);
   WriteDbl(s, "ALTRPIX ", f->altrpix);
   WriteDbl(s, "OBSRA   ", f->obsra);
   WriteDbl(s, "OBSDEC  ", f->obsdec);
   WriteDbl(s, "RESTFREQ", f->restfreq);
   WriteDbl(s, "YSHIFT  ", f->yshift);
   WriteDbl(s, "DATAMAX ", f->datamax);
   WriteDbl(s, "DATAMIN ", f->datamin);
   WriteStr(s, "CTYPE1  ", f->ctype1);
   WriteDbl(s, "CRVAL1  ", f->crval1);
   WriteDbl(s, "CDELT1  ", f->cdelt1);
   WriteDbl(s, "CRPIX1  ", f->crpix1);
   WriteDbl(s, "CROTA1  ", f->crota1);
   if (f->naxis >= 2) {
      WriteStr(s, "CTYPE2  ", f->ctype2);
      WriteDbl(s, "CRVAL2  ", f->crval2);
      WriteDbl(s, "CDELT2  ", f->cdelt2);
      WriteDbl(s, "CRPIX2  ", f->crpix2);
      WriteDbl(s, "CROTA2  ", f->crota2);
   }
   if (f->naxis >= 3) {
      WriteStr(s, "CTYPE3  ", f->ctype3);
      WriteDbl(s, "CRVAL3  ", f->crval3);
      WriteDbl(s, "CDELT3  ", f->cdelt3);
      WriteDbl(s, "CRPIX3  ", f->crpix3);
      WriteDbl(s, "CROTA3  ", f->crota3);
   }
   if (f->naxis >= 4) {
      WriteStr(s, "CTYPE4  ", f->ctype4);
      WriteDbl(s, "CRVAL4  ", f->crval4);
      WriteDbl(s, "CDELT4  ", f->cdelt4);
      WriteDbl(s, "CRPIX4  ", f->crpix4);
      WriteDbl(s, "CROTA4  ", f->crota4);
   }
   write(outfile, "END                                                                             ", 80); nline++;

   for (; nline % 36 != 0; nline++)
      write(outfile, SPC80, 80);

} /* WriteImHeader() */

void WriteSpHeader(outfile, f)
/* write FITS (spectrum) header by newstar style. 93/06/13 I.Yamamura */
   int outfile;
   fits_header *f;
{
   char s[81], ps[9];
   int nline = 0;
   int i;

   WriteBln(s, "SIMPLE  ", f->simple);
   WriteInt(s, "BITPIX  ", f->bitpix);
   WriteInt(s, "NAXIS   ", f->naxis);
   WriteInt(s, "NAXIS1  ", f->naxis1);
   if (f->naxis >= 2) WriteInt(s, "NAXIS2  ", f->naxis2);
   if (f->naxis >= 3) WriteInt(s, "NAXIS3  ", f->naxis3);
   if (f->naxis >= 4) WriteInt(s, "NAXIS4  ", f->naxis4);
   WriteBln(s, "GROUPS  ", f->groups);
   WriteInt(s, "PCOUNT  ", f->pcount);
   WriteInt(s, "GCOUNT  ", f->gcount);
   if (f->groups == 'T') {
      for (i = 0; i < f->pcount; i++) {
         sprintf(ps, "PTYPE%-3d", i+1);
         WriteStr(s, ps, (f->p[i]).type);
      }
      for (i = 0; i < f->pcount; i++) {
         sprintf(ps, "PSCAL%-3d", i+1);
         WriteDbl(s, ps, (f->p[i]).scal);
      }
   }
   if (f->naxis1 > 0) {
      WriteStr(s, "CTYPE1  ", f->ctype1);
      WriteDbl(s, "CRVAL1  ", f->crval1);
      WriteDbl(s, "CDELT1  ", f->cdelt1);
      WriteDbl(s, "CRPIX1  ", f->crpix1);
      WriteDbl(s, "CROTA1  ", f->crota1);
   }
   if (f->naxis >= 2) {
      WriteStr(s, "CTYPE2  ", f->ctype2);
      WriteDbl(s, "CRVAL2  ", f->crval2);
      WriteDbl(s, "CDELT2  ", f->cdelt2);
      WriteDbl(s, "CRPIX2  ", f->crpix2);
      WriteDbl(s, "CROTA2  ", f->crota2);
   }
   if (f->naxis >= 3) {
      WriteStr(s, "CTYPE3  ", f->ctype3);
      WriteDbl(s, "CRVAL3  ", f->crval3);
      WriteDbl(s, "CDELT3  ", f->cdelt3);
      WriteDbl(s, "CRPIX3  ", f->crpix3);
      WriteDbl(s, "CROTA3  ", f->crota3);
   }
   if (f->naxis >= 4) {
      WriteStr(s, "CTYPE4  ", f->ctype4);
      WriteDbl(s, "CRVAL4  ", f->crval4);
      WriteDbl(s, "CDELT4  ", f->cdelt4);
      WriteDbl(s, "CRPIX4  ", f->crpix4);
      WriteDbl(s, "CROTA4  ", f->crota4);
   }
   WriteDbl(s, "BSCALE  ", f->bscale);
   WriteDbl(s, "BZERO   ", f->bzero);
   WriteStr(s, "BUNIT   ", f->bunit);
   WriteInt(s, "BLANK   ", f->blank);
   WriteStr(s, "OBJECT  ", f->object);
   WriteStr(s, "ORIGIN  ", f->origin);
   WriteStr(s, "INSTRUME", f->instrume);
   WriteStr(s, "TELESCOP", f->telescop);
   WriteStr(s, "OBSERVER", f->observer);
   WriteDbl(s, "EPOCH   ", f->epoch);
   write(outfile, "END                                                                             ", 80); nline++;

   for (; nline % 36 != 0; nline++)
      write(outfile, SPC80, 80);

} /* WriteSpHeader() */

void WriteUVHeader(outfile, f)
/* write FITS (UV) header by AIPS style. 95/03/11 I.Yamamura */
   int outfile;
   fits_header *f;
{
   char s[81], ps[9];
   int nline = 0;
   int i;

   WriteBln(s, "SIMPLE  ", f->simple);
   WriteInt(s, "BITPIX  ", f->bitpix);
   WriteInt(s, "NAXIS   ", f->naxis);
   WriteInt(s, "NAXIS1  ", 0);
   if (f->naxis >= 2) WriteInt(s, "NAXIS2  ", f->naxis2);
   if (f->naxis >= 3) WriteInt(s, "NAXIS3  ", f->naxis3);
   if (f->naxis >= 4) WriteInt(s, "NAXIS4  ", f->naxis4);
   if (f->naxis >= 5) WriteInt(s, "NAXIS5  ", f->naxis5);
   if (f->naxis >= 6) WriteInt(s, "NAXIS6  ", f->naxis6);
   WriteBln(s, "EXTEND  ", f->extend);
   WriteBln(s, "BLOCKED ", f->blocked);
   WriteStr(s, "OBJECT  ", f->object);
   WriteStr(s, "TELESCOP", f->telescop);
   WriteStr(s, "INSTRUME", f->instrume);
   WriteStr(s, "OBSERVER", f->observer);
   WriteStr(s, "DATE-OBS", f->date_obs);
   WriteStr(s, "DATE-MAP", f->date_map);
   WriteDbl(s, "BSCALE  ", f->bscale);
   WriteDbl(s, "BZERO   ", f->bzero);
   WriteStr(s, "BUNIT   ", f->bunit);
   WriteDbl(s, "EPOCH   ", f->epoch);
   WriteInt(s, "BLANK   ", f->blank);
   WriteDbl(s, "ALTRVAL ", f->altrval);
   WriteDbl(s, "OBSRA   ", f->obsra);
   WriteDbl(s, "OBSDEC  ", f->obsdec);
   if (f->naxis >= 2) {
      WriteStr(s, "CTYPE2  ", f->ctype2);
      WriteDbl(s, "CRVAL2  ", f->crval2);
      WriteDbl(s, "CDELT2  ", f->cdelt2);
      WriteDbl(s, "CRPIX2  ", f->crpix2);
      WriteDbl(s, "CROTA2  ", f->crota2);
   }
   if (f->naxis >= 3) {
      WriteStr(s, "CTYPE3  ", f->ctype3);
      WriteDbl(s, "CRVAL3  ", f->crval3);
      WriteDbl(s, "CDELT3  ", f->cdelt3);
      WriteDbl(s, "CRPIX3  ", f->crpix3);
      WriteDbl(s, "CROTA3  ", f->crota3);
   }
   if (f->naxis >= 4) {
      WriteStr(s, "CTYPE4  ", f->ctype4);
      WriteDbl(s, "CRVAL4  ", f->crval4);
      WriteDbl(s, "CDELT4  ", f->cdelt4);
      WriteDbl(s, "CRPIX4  ", f->crpix4);
      WriteDbl(s, "CROTA4  ", f->crota4);
   }
   if (f->naxis >= 5) {
      WriteStr(s, "CTYPE5  ", f->ctype5);
      WriteDbl(s, "CRVAL5  ", f->crval5);
      WriteDbl(s, "CDELT5  ", f->cdelt5);
      WriteDbl(s, "CRPIX5  ", f->crpix5);
      WriteDbl(s, "CROTA5  ", f->crota5);
   }
   if (f->naxis >= 6) {
      WriteStr(s, "CTYPE6  ", f->ctype6);
      WriteDbl(s, "CRVAL6  ", f->crval6);
      WriteDbl(s, "CDELT6  ", f->cdelt6);
      WriteDbl(s, "CRPIX6  ", f->crpix6);
      WriteDbl(s, "CROTA6  ", f->crota6);
   }
   WriteBln(s, "GROUPS  ", f->groups);
   WriteInt(s, "GCOUNT  ", f->gcount);
   WriteInt(s, "PCOUNT  ", f->pcount);
   if (f->groups == 'T') {
      for (i = 0; i < f->pcount; i++) {
         sprintf(ps, "PTYPE%-3d", i+1);
         WriteStr(s, ps, (f->p[i]).type);
         sprintf(ps, "PSCAL%-3d", i+1);
         WriteDbl(s, ps, (f->p[i]).scal);
         sprintf(ps, "PZERO%-3d", i+1);
         WriteDbl(s, ps, (f->p[i]).zero);
      }
   }
   WriteStr(s, "ORIGIN  ", f->origin);
   sprintf(s, "HISTORY AIPS WTSCAL = %18.11E / CMPLX WTS=WTSCAL*(TAPE*BSCALE+BZERO) ", f->wtscal);
   write(outfile, s, 80); nline++;
   write(outfile, "END                                                                             ", 80); nline++;

   for (; nline % 36 != 0; nline++)
      write(outfile, SPC80, 80);

} /* WriteUVHeader() */

void WriteImage(fn, img, f, mode)
/* write image data to fits file */
/* force to write if mode > 0. 93/03/21 */
/* 93/09/07 recalculate size of the element before saving */
   char *fn;
   long *img;
   fits_header *f;
   int mode;
{
   int outfile;
   int iv, ira, idec;
   unsigned long size;
   char ans;
   void *spc;     /* fill the file to make the size to be 2880 * n */

   outfile = open(fn, (O_CREAT | O_EXCL | O_WRONLY), 0644);
   if (outfile == -1) {
      if (errno == EEXIST) {
         if (mode == 0) {
            fprintf(stderr, "WriteImage: File %s already exists. Overwrite(y/n)? ", fn);
            scanf("%1s", &ans);
            if ((ans == 'y') || (ans == 'Y'))
               outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
            else
               exit(0);
         } else
            outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
      }
      if (outfile == -1) {
         fprintf(stderr, "WriteImage: File %s opening error!\n", fn);
         exit(-1);
      }
   }
   WriteImHeader(outfile, f);

   if (f->naxis >= 1) f->size  = f->naxis1;
   if (f->naxis >= 2) f->size *= f->naxis2;
   if (f->naxis >= 3) f->size *= f->naxis3;
   if (f->naxis >= 4) f->size *= f->naxis4;
   write(outfile, img, f->size * sizeof(long));
   spc = (void *)calloc((2880 - (f->size * sizeof(long)) % 2880), 1);
   write(outfile, spc, (2880 - (f->size * sizeof(long)) % 2880));

   close(outfile);
} /* WriteImage() */

void WriteSpectra(fn, spc, f, mode)
/* write spectral data to fits file 93/06/13 */
/* 93/09/07 recalculate size of the element before saving */
   char *fn;
   long *spc;
   fits_header *f;
   int mode;
{
   int outfile;
   int iv, ira, idec;
   unsigned long size;
   char ans;

   outfile = open(fn, (O_CREAT | O_EXCL | O_WRONLY), 0644);
   if (outfile == -1) {
      if (errno == EEXIST) {
         if (mode == 0) {
            fprintf(stderr, "WriteSpectra: File %s already exists. Overwrite(y/n)? ", fn);
            scanf("%1s", &ans);
            if ((ans == 'y') || (ans == 'Y'))
               outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
            else
               exit(0);
         } else
            outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
      }
      if (outfile == -1) {
         fprintf(stderr, "WriteSpectra: File %s opening error!\n", fn);
         exit(-1);
      }
   }
   WriteSpHeader(outfile, f);

   f->size = (f->naxis2 + f->pcount) * f->gcount;
   write(outfile, spc, f->size * sizeof(long));

   close(outfile);

} /* WriteSpectra() */

/* ----------------------------------------------------------------------- */
void WriteUV(fn, fuv, f, mode)
/* write UV data to fits file 93/06/13 */
/* 95/03/11 */
   char *fn;
   int *fuv;        /* pointer for the data */
   fits_header *f;
   int mode;
{
   int outfile;
   char ans;

   outfile = open(fn, (O_CREAT | O_EXCL | O_WRONLY), 0644);
   if (outfile == -1) {
      if (errno == EEXIST) {
         if (mode == 0) {
            fprintf(stderr, "WriteUV: File %s already exists. Overwrite(y/n)? ", fn);
            scanf("%1s", &ans);
            if ((ans == 'y') || (ans == 'Y'))
               outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
            else
               exit(0);
         } else
            outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
      }
      if (outfile == -1) {
         fprintf(stderr, "WriteUV: File %s opening error!\n", fn);
         exit(-1);
      }
   }
   WriteUVHeader(outfile, f);
   *fuv = outfile;

} /* WriteUV() */

/* ======================================================================== */
void WriteASCII(asfn, img, f, format, force)
/* save to ascii file named asfn */
/* 93/06/14 add format option, 4-d data can be treated */
/*          moved from imsub.c */
   long *img;
   fits_header *f;
   char *asfn;
   int format; /* writeing format (0: integer / 1: float) */
   int force;  /* force to write (if 1)*/
{
#define TBX (img[((((iv)*f->naxis2)+idec)*f->naxis1+ira)]*f->bscale+f->bzero)
   FILE *af;
   int io, iv, idec, ira;
   int iomax, ivmax, idecmax, iramax;
   int ibx;
   char ans;

/* check if the same file exists */
   if (force == 1)   /* force overwrite mode */
      ans = 'y';
   else {
      af = fopen(asfn, "r");
      if (af != NULL) {
         ans = 'n';
         fprintf(stderr, "WriteASCII: File %s already exists. Overwrite(y/n)? ", asfn);
         scanf("%1s", &ans);
         fclose(af);
      } else
         ans = 'y';
   }

/* open files */
   if ((ans == 'y') || (ans == 'Y')) {
      af = fopen(asfn, "w");
      if (af == NULL) {
         fprintf(stderr, "WriteASCII: File open error.\n");
         exit(-1);
      }
   } else
      exit(0);

   if (f->naxis >= 1) iramax =  f->naxis1; else iramax = 1;
   if (f->naxis >= 2) idecmax = f->naxis2; else idecmax = 1;
   if (f->naxis >= 3) ivmax =   f->naxis3; else ivmax = 1;
   if (f->naxis >= 4) iomax =   f->naxis4; else iomax = 1;

   for (io = 0; io < iomax; io++) {
      for (iv = 0; iv < ivmax; iv++) {
         for (idec = 0; idec < idecmax; idec++) {
            for (ira = 0; ira < iramax; ira++) {
               if (format == 0) { /* written in integer between 0 to 255 */
                  ibx = (int)((TBX - f->datamin) /
                        (f->datamax - f->datamin) * 256);
                  fprintf(af, "%4d ",ibx);
               } else /* writen in float */
                  fprintf(af, "%7.2f ",TBX);
            }
            fprintf(af, "\n");
         }
      }
   }
   fclose(af);
} /* WriteASCII() */

/* ======================================================================== */
void WriteFbin(fn, img, f, mode)
/* save to floating binary (6-bytes) file named asfn */
/* 93/08/02 by I.Yamamura */
   char *fn;
   long *img;
   fits_header *f;
   int mode;
{
   int outfile;
   int i, iv, ira, idec;
   unsigned long size;
   char ans;
   float *fimg;

   outfile = open(fn, (O_CREAT | O_EXCL | O_WRONLY), 0644);
   if (outfile == -1) {
      if (errno == EEXIST) {
         if (mode == 0) {
            fprintf(stderr, "WriteImage: File %s already exists. Overwrite(y/n)? ", fn);
            scanf("%1s", &ans);
            if ((ans == 'y') || (ans == 'Y'))
               outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
            else
               exit(0);
         } else
            outfile = open(fn, (O_TRUNC | O_WRONLY), 0);
      }
      if (outfile == -1) {
         fprintf(stderr, "WriteImage: File opening error!\n");
         exit(-1);
      }
   }

   fimg = (float *)calloc(f->size, sizeof(float));
   for (i = 0; i < f->size; i++)
      fimg[i] = (float)(img[i] * f->bscale + f->bzero);

   write(outfile, fimg, f->size * sizeof(float));

   close(outfile);
} /* WriteFbin() */

/* end of rwfits.c */
